module PetscHeaders
#   define PETSC_USE_FORTRAN_INTERFACES 1
#   include <finclude/petsc.h90>
#   undef PETSC_USE_FORTRAN_INTERFACES 
end module PetscHeaders

module PetscMPI
use iso_c_binding

     logical :: events_initialized = .false.
     integer :: event_Wait, event_Barrier, event_iSend, event_Send, event_Recv, event_Reduce, event_Allreduce, event_Allgather

interface  

     function PetscAdminSend(count, datatype) bind(c,name='PetscAdminSend') result(ierr)
     use iso_c_binding
     implicit none
     integer(c_int), value :: count
     integer(c_int), value :: datatype
     integer(c_int)        :: ierr
     end function PetscAdminSend

     function PetscAdminRecv(count, datatype) bind(c,name='PetscAdminRecv') result(ierr)
     use iso_c_binding
     implicit none
     integer(c_int), value :: count
     integer(c_int), value :: datatype
     integer(c_int)        :: ierr
     end function PetscAdminRecv

     function PetscAdminReduce() bind(c,name='PetscAdminReduce') result(ierr)
     use iso_c_binding
     implicit none
     integer(c_int)        :: ierr
     end function PetscAdminReduce

end interface 

interface PetscMPI_iSend
    module procedure PetscMPI_iSend_Real, PetscMPI_iSend_Double, PetscMPI_iSend_Integer, &
                    PetscMPI_iSend_Real2, PetscMPI_iSend_Double2, PetscMPI_iSend_Integer2, &
                    PetscMPI_iSend_Real3, PetscMPI_iSend_Double3, PetscMPI_iSend_Integer3
end interface PetscMPI_iSend

interface PetscMPI_Send
    module procedure PetscMPI_Send_Real, PetscMPI_Send_Double, PetscMPI_Send_Integer, &
                    PetscMPI_Send_Real2, PetscMPI_Send_Double2, PetscMPI_Send_Integer2, &
                    PetscMPI_Send_Real3, PetscMPI_Send_Double3, PetscMPI_Send_Integer3
end interface PetscMPI_Send

interface PetscMPI_Recv
    module procedure PetscMPI_Recv_Real, PetscMPI_Recv_Double, PetscMPI_Recv_Integer, &
                    PetscMPI_Recv_Real2, PetscMPI_Recv_Double2, PetscMPI_Recv_Integer2, &
                    PetscMPI_Recv_Real3, PetscMPI_Recv_Double3, PetscMPI_Recv_Integer3
end interface PetscMPI_Recv

interface PetscMPI_Reduce
    module procedure PetscMPI_Reduce_Real, PetscMPI_Reduce_Double, PetscMPI_Reduce_Integer, &
                    PetscMPI_Reduce_Real0, PetscMPI_Reduce_Double0, PetscMPI_Reduce_Integer0, &
                    PetscMPI_Reduce_Real2, PetscMPI_Reduce_Double2, PetscMPI_Reduce_Integer2, &
                    PetscMPI_Reduce_Real3, PetscMPI_Reduce_Double3, PetscMPI_Reduce_Integer3
end interface PetscMPI_Reduce

interface PetscMPI_AllReduce
    module procedure PetscMPI_Allreduce_Real, PetscMPI_Allreduce_Double, PetscMPI_Allreduce_Integer, &
                    PetscMPI_Allreduce_Real0, PetscMPI_Allreduce_Double0, PetscMPI_Allreduce_Integer0, &
                    PetscMPI_Allreduce_Real2, PetscMPI_Allreduce_Double2, PetscMPI_Allreduce_Integer2, &
                    PetscMPI_Allreduce_Real3, PetscMPI_Allreduce_Double3, PetscMPI_Allreduce_Integer3
end interface PetscMPI_AllReduce

interface PetscMPI_Allgather
    module procedure PetscMPI_Allgather_Integer, PetscMPI_Allgather_Integer0 
                    ! PetscMPI_Allgather_Real, PetscMPI_Allgather_Double, 
                    ! , &
                    ! PetscMPI_Allgather_Real0, PetscMPI_Allgather_Double0
                    ! PetscMPI_Allgather_Real2, PetscMPI_Allgather_Double2, PetscMPI_Allgather_Integer2, &
                    ! PetscMPI_Allgather_Real3, PetscMPI_Allgather_Double3, PetscMPI_Allgather_Integer3
end interface PetscMPI_AllGather

contains
     subroutine PetscMPI_initialize_events(ierr)
     use myLogEventF
     use petsc
     implicit none
        integer, intent(out) :: ierr
        integer :: rank
        if (events_initialized) then
           call MPI_Comm_rank(PETSC_COMM_WORLD, rank,ierr)
           print '(a,i0,a)','[',rank,'] ERROR: Performance tracking events have already been initialized.'
           ierr = PETSC_ERR_ARG_INCOMP 
           return
        end if
        call PetscLogEventRegister('MPI_Wait',     0, event_Wait,      ierr)
        call PetscLogEventRegister('MPI_Barrier',  0, event_Barrier,   ierr)
        call PetscLogEventRegister('MPI_iSend',    0, event_iSend,     ierr)
        call PetscLogEventRegister('MPI_Send',     0, event_Send,      ierr)
        call PetscLogEventRegister('MPI_Recv',     0, event_Recv,      ierr)
        call PetscLogEventRegister('MPI_Reduce',   0, event_Reduce,    ierr)
        call PetscLogEventRegister('MPI_AllReduce',0, event_Allreduce, ierr)
        call PetscLogEventRegister('MPI_AllGather',0, event_AllGather, ierr)
        events_initialized = .true.
     end subroutine PetscMPI_initialize_events

     subroutine PetscMPI_Barrier(comm, ierr)
     use petsc
     implicit none
         integer, intent(in)  :: comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Barrier,ierr)
         call MPI_Barrier(comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Barrier,ierr)
     end subroutine PetscMPI_Barrier

     subroutine PetscMPI_Wait(req, status, ierr)
     use petsc
     implicit none
         integer              :: req
         integer              :: status(:)
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Wait,ierr)
         call MPI_Wait(req, status, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Wait,ierr)
     end subroutine PetscMPI_Wait

     subroutine PetscMPI_iSend_Real( message, msgLen, datatype, dest, tag, comm, req, ierr)
     use petsc
     implicit none
         real :: message(:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: req
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_iSend,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_iSend( message, msgLen, datatype, dest, tag, comm, req, ierr)
         if (events_initialized) call PetscLogEventEnd(event_iSend,ierr)
     end subroutine PetscMPI_iSend_Real

     subroutine PetscMPI_iSend_Double( message, msgLen, datatype, dest, tag, comm, req, ierr)
     implicit none
         double precision :: message(:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: req
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_iSend,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_iSend( message, msgLen, datatype, dest, tag, comm, req, ierr)
         if (events_initialized) call PetscLogEventEnd(event_iSend,ierr)
     end subroutine PetscMPI_iSend_Double

     subroutine PetscMPI_iSend_Integer( message, msgLen, datatype, dest, tag, comm, req, ierr)
     implicit none
         integer :: message(:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: req
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_iSend,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_iSend( message, msgLen, datatype, dest, tag, comm, req, ierr)
         if (events_initialized) call PetscLogEventEnd(event_iSend,ierr)
     end subroutine PetscMPI_iSend_Integer

     subroutine PetscMPI_iSend_Real2( message, msgLen, datatype, dest, tag, comm, req, ierr)
     implicit none
         real :: message(:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: req
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_iSend,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_iSend( message, msgLen, datatype, dest, tag, comm, req, ierr)
         if (events_initialized) call PetscLogEventEnd(event_iSend,ierr)
     end subroutine PetscMPI_iSend_Real2

     subroutine PetscMPI_iSend_Double2( message, msgLen, datatype, dest, tag, comm, req, ierr)
     implicit none
         double precision :: message(:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: req
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_iSend,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_iSend( message, msgLen, datatype, dest, tag, comm, req, ierr)
         if (events_initialized) call PetscLogEventEnd(event_iSend,ierr)
     end subroutine PetscMPI_iSend_Double2

     subroutine PetscMPI_iSend_Integer2( message, msgLen, datatype, dest, tag, comm, req, ierr)
     implicit none
         integer :: message(:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: req
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_iSend,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_iSend( message, msgLen, datatype, dest, tag, comm, req, ierr)
         if (events_initialized) call PetscLogEventEnd(event_iSend,ierr)
     end subroutine PetscMPI_iSend_Integer2

     subroutine PetscMPI_iSend_Real3( message, msgLen, datatype, dest, tag, comm, req, ierr)
     implicit none
         real :: message(:,:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: req
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_iSend,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_iSend( message, msgLen, datatype, dest, tag, comm, req, ierr)
         if (events_initialized) call PetscLogEventEnd(event_iSend,ierr)
     end subroutine PetscMPI_iSend_Real3

     subroutine PetscMPI_iSend_Double3( message, msgLen, datatype, dest, tag, comm, req, ierr)
     implicit none
         double precision :: message(:,:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: req
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_iSend,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_iSend( message, msgLen, datatype, dest, tag, comm, req, ierr)
         if (events_initialized) call PetscLogEventEnd(event_iSend,ierr)
     end subroutine PetscMPI_iSend_Double3

     subroutine PetscMPI_iSend_Integer3( message, msgLen, datatype, dest, tag, comm, req, ierr)
     implicit none
         integer :: message(:,:,:)
         integer, intent(in)  :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: req
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_iSend,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_iSend( message, msgLen, datatype, dest, tag, comm, req, ierr)
         if (events_initialized) call PetscLogEventEnd(event_iSend,ierr)
     end subroutine PetscMPI_iSend_Integer3

     !============================= Send =======================================
     subroutine PetscMPI_Send_Real( message, msgLen, datatype, dest, tag, comm, ierr)
     implicit none
         real :: message(:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Send,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_Send( message, msgLen, datatype, dest, tag, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Send,ierr)
     end subroutine PetscMPI_Send_Real

     subroutine PetscMPI_Send_Double( message, msgLen, datatype, dest, tag, comm, ierr)
     implicit none
         double precision :: message(:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Send,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_Send( message, msgLen, datatype, dest, tag, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Send,ierr)
     end subroutine PetscMPI_Send_Double

     subroutine PetscMPI_Send_Integer( message, msgLen, datatype, dest, tag, comm, ierr)
     implicit none
         integer :: message(:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Send,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_Send( message, msgLen, datatype, dest, tag, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Send,ierr)
     end subroutine PetscMPI_Send_Integer

     subroutine PetscMPI_Send_Real2( message, msgLen, datatype, dest, tag, comm, ierr)
     implicit none
         real :: message(:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Send,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_Send( message, msgLen, datatype, dest, tag, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Send,ierr)
     end subroutine PetscMPI_Send_Real2

     subroutine PetscMPI_Send_Double2( message, msgLen, datatype, dest, tag, comm, ierr)
     implicit none
         double precision :: message(:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Send,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_Send( message, msgLen, datatype, dest, tag, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Send,ierr)
     end subroutine PetscMPI_Send_Double2

     subroutine PetscMPI_Send_Integer2( message, msgLen, datatype, dest, tag, comm, ierr)
     implicit none
         integer :: message(:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Send,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_Send( message, msgLen, datatype, dest, tag, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Send,ierr)
     end subroutine PetscMPI_Send_Integer2

     subroutine PetscMPI_Send_Real3( message, msgLen, datatype, dest, tag, comm, ierr)
     implicit none
         real :: message(:,:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Send,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_Send( message, msgLen, datatype, dest, tag, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Send,ierr)
     end subroutine PetscMPI_Send_Real3

     subroutine PetscMPI_Send_Double3( message, msgLen, datatype, dest, tag, comm, ierr)
     implicit none
         double precision :: message(:,:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Send,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_Send( message, msgLen, datatype, dest, tag, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Send,ierr)
     end subroutine PetscMPI_Send_Double3

     subroutine PetscMPI_Send_Integer3( message, msgLen, datatype, dest, tag, comm, ierr)
     implicit none
         integer :: message(:,:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Send,ierr)
         ierr = PetscAdminSend( msgLen, datatype)
         call MPI_Send( message, msgLen, datatype, dest, tag, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Send,ierr)
     end subroutine PetscMPI_Send_Integer3



     !============================= Receive =======================================
     subroutine PetscMPI_Recv_Real( message, msgLen, datatype, dest, tag, comm, status, ierr)
     implicit none
         real :: message(:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer             :: status(:)
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Recv,ierr)
         ierr = PetscAdminRecv( msgLen, datatype)
         call MPI_Recv( message, msgLen, datatype, dest, tag, comm, status, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Recv,ierr)
     end subroutine PetscMPI_Recv_Real

     subroutine PetscMPI_Recv_Real2( message, msgLen, datatype, dest, tag, comm, status, ierr)
     implicit none
         real :: message(:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer             :: status(:)
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Recv,ierr)
         ierr = PetscAdminRecv( msgLen, datatype)
         call MPI_Recv( message, msgLen, datatype, dest, tag, comm, status, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Recv,ierr)
     end subroutine PetscMPI_Recv_Real2

     subroutine PetscMPI_Recv_Real3( message, msgLen, datatype, dest, tag, comm, status, ierr)
     implicit none
         real :: message(:,:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer             :: status(:)
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Recv,ierr)
         ierr = PetscAdminRecv( msgLen, datatype)
         call MPI_Recv( message, msgLen, datatype, dest, tag, comm, status, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Recv,ierr)
     end subroutine PetscMPI_Recv_Real3

     subroutine PetscMPI_Recv_Integer( message, msgLen, datatype, dest, tag, comm, status, ierr)
     implicit none
         integer :: message(:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer             :: status(:)
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Recv,ierr)
         ierr = PetscAdminRecv( msgLen, datatype)
         call MPI_Recv( message, msgLen, datatype, dest, tag, comm, status, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Recv,ierr)
     end subroutine PetscMPI_Recv_Integer

     subroutine PetscMPI_Recv_Integer2( message, msgLen, datatype, dest, tag, comm, status, ierr)
     implicit none
         integer :: message(:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer             :: status(:)
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Recv,ierr)
         ierr = PetscAdminRecv( msgLen, datatype)
         call MPI_Recv( message, msgLen, datatype, dest, tag, comm, status, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Recv,ierr)
     end subroutine PetscMPI_Recv_Integer2

     subroutine PetscMPI_Recv_Integer3( message, msgLen, datatype, dest, tag, comm, status, ierr)
     implicit none
         integer :: message(:,:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer             :: status(:)
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Recv,ierr)
         ierr = PetscAdminRecv( msgLen, datatype)
         call MPI_Recv( message, msgLen, datatype, dest, tag, comm, status, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Recv,ierr)
     end subroutine PetscMPI_Recv_Integer3

     subroutine PetscMPI_Recv_Double( message, msgLen, datatype, dest, tag, comm, status, ierr)
     implicit none
         double precision :: message(:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer             :: status(:)
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Recv,ierr)
         ierr = PetscAdminRecv( msgLen, datatype)
         call MPI_Recv( message, msgLen, datatype, dest, tag, comm, status, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Recv,ierr)
     end subroutine PetscMPI_Recv_Double

     subroutine PetscMPI_Recv_Double2( message, msgLen, datatype, dest, tag, comm, status, ierr)
     implicit none
         double precision :: message(:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer             :: status(:)
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Recv,ierr)
         ierr = PetscAdminRecv( msgLen, datatype)
         call MPI_Recv( message, msgLen, datatype, dest, tag, comm, status, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Recv,ierr)
     end subroutine PetscMPI_Recv_Double2

     subroutine PetscMPI_Recv_Double3( message, msgLen, datatype, dest, tag, comm, status, ierr)
     implicit none
         double precision :: message(:,:,:)
         integer, intent(in) :: msgLen, datatype, dest, tag, comm
         integer             :: status(:)
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Recv,ierr)
         ierr = PetscAdminRecv( msgLen, datatype)
         call MPI_Recv( message, msgLen, datatype, dest, tag, comm, status, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Recv,ierr)
     end subroutine PetscMPI_Recv_Double3


     !============================= Reduce =======================================
     subroutine PetscMPI_Reduce_Real0( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         real :: mess1, mess2
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Real0

     subroutine PetscMPI_Reduce_Real( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         real :: mess1(:), mess2(:)
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Real

     subroutine PetscMPI_Reduce_Real2( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         real :: mess1(:,:), mess2(:,:)
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Real2

     subroutine PetscMPI_Reduce_Real3( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         real :: mess1(:,:,:), mess2(:,:,:)
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Real3

     subroutine PetscMPI_Reduce_Integer0( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         integer :: mess1, mess2
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Integer0

     subroutine PetscMPI_Reduce_Integer( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         integer :: mess1(:), mess2(:)
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Integer

     subroutine PetscMPI_Reduce_Integer2( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         integer :: mess1(:,:), mess2(:,:)
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Integer2

     subroutine PetscMPI_Reduce_Integer3( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         integer :: mess1(:,:,:), mess2(:,:,:)
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Integer3

     subroutine PetscMPI_Reduce_Double0( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         double precision :: mess1, mess2
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Double0

     subroutine PetscMPI_Reduce_Double( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         double precision :: mess1(:), mess2(:)
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Double

     subroutine PetscMPI_Reduce_Double2( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         double precision :: mess1(:,:), mess2(:,:)
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Double2

     subroutine PetscMPI_Reduce_Double3( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
     implicit none
         double precision :: mess1(:,:,:), mess2(:,:,:)
         integer, intent(in) :: msgLen, datatype, op, root, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_Reduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Reduce( mess1, mess2, msgLen, datatype, op, root, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_Reduce,ierr)
     end subroutine PetscMPI_Reduce_Double3

     !============================= Allreduce =======================================
     subroutine PetscMPI_Allreduce_Real0( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         real :: mess1, mess2
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Real0

     subroutine PetscMPI_Allreduce_Real( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         real :: mess1(:), mess2(:)
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Real

     subroutine PetscMPI_Allreduce_Real2( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         real :: mess1(:,:), mess2(:,:)
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Real2

     subroutine PetscMPI_Allreduce_Real3( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         real :: mess1(:,:,:), mess2(:,:,:)
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Real3

     subroutine PetscMPI_Allreduce_Integer0( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         integer :: mess1, mess2
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Integer0

     subroutine PetscMPI_Allreduce_Integer( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         integer :: mess1(:), mess2(:)
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Integer

     subroutine PetscMPI_Allreduce_Integer2( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         integer :: mess1(:,:), mess2(:,:)
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Integer2

     subroutine PetscMPI_Allreduce_Integer3( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         integer :: mess1(:,:,:), mess2(:,:,:)
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Integer3

     subroutine PetscMPI_Allreduce_Double0( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         double precision :: mess1, mess2
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Double0

     subroutine PetscMPI_Allreduce_Double( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         double precision :: mess1(:), mess2(:)
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Double

     subroutine PetscMPI_Allreduce_Double2( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         double precision :: mess1(:,:), mess2(:,:)
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Double2

     subroutine PetscMPI_Allreduce_Double3( mess1, mess2, msgLen, datatype, op, comm, ierr)
     implicit none
         double precision :: mess1(:,:,:), mess2(:,:,:)
         integer, intent(in) :: msgLen, datatype, op, comm
         integer, intent(out) :: ierr
         if (events_initialized) call PetscLogEventBegin(event_AllReduce,ierr)
         ierr = PetscAdminReduce()
         call MPI_Allreduce( mess1, mess2, msgLen, datatype, op, comm, ierr)
         if (events_initialized) call PetscLogEventEnd(event_AllReduce,ierr)
     end subroutine PetscMPI_Allreduce_Double3

     !============================= AllGather =======================================
     subroutine PetscMPI_Allgather_Integer( send, sendcount, sendtype, receive, recvcount, recvtype, comm, ierr)
     use petsc
     implicit none
         integer, intent(in)  :: send(:)
         integer, intent(in)  :: sendcount, sendtype
         integer, intent(out) :: receive(:,:)
         integer, intent(in)  :: recvcount, recvtype
         integer, intent(in)  :: comm
         integer, intent(out) :: ierr
         integer :: nprocs, rank
         if (events_initialized) call PetscLogEventBegin(event_AllGather,ierr)
         if (recvcount /= sendcount .or. recvtype /= sendtype) then
            call MPI_Comm_rank(PETSC_COMM_WORLD, rank,ierr)
            print '(a,i0,a)','[',rank,'] ERROR: Dimension/type of send and receive buffer do not correspond.'
            ierr = PETSC_ERR_ARG_INCOMP 
            return
         end if
         call MPI_Allgather( send, sendcount, sendtype, receive, recvcount, recvtype, comm, ierr)
         call MPI_Comm_size( comm, nprocs, ierr)

         ierr = PetscAdminSend( sendcount * (nprocs-1), sendtype)
         ierr = PetscAdminRecv( sendcount * (nprocs-1), sendtype)
         if (events_initialized) call PetscLogEventEnd(event_AllGather,ierr)
     end subroutine PetscMPI_Allgather_Integer

     subroutine PetscMPI_Allgather_Integer0( send, sendcount, sendtype, receive, recvcount, recvtype, comm, ierr)
     use petsc
     implicit none
         integer, intent(in)  :: send
         integer, intent(in)  :: sendcount, sendtype
         integer, intent(out) :: receive(:)
         integer, intent(in)  :: recvcount, recvtype
         integer, intent(in)  :: comm
         integer, intent(out) :: ierr
         integer :: nprocs, rank
         if (events_initialized) call PetscLogEventBegin(event_AllGather,ierr)
         if (recvcount /= 1 .or. sendcount/=1 .or. recvtype /= sendtype) then
            call MPI_Comm_rank(PETSC_COMM_WORLD, rank,ierr)
            print '(a,i0,a)','[',rank,'] ERROR: Dimension/type of send and receive buffer do not correspond.'
            ierr = PETSC_ERR_ARG_INCOMP 
            return
         end if
         call MPI_Allgather( send, sendcount, sendtype, receive, recvcount, recvtype, comm, ierr)
         call MPI_Comm_size( comm, nprocs, ierr)

         ierr = PetscAdminSend( nprocs-1, sendtype)
         ierr = PetscAdminRecv( nprocs-1, sendtype)
         if (events_initialized) call PetscLogEventEnd(event_AllGather,ierr)
     end subroutine PetscMPI_Allgather_Integer0

end module PetscMPI

