module XMLViewerF
use iso_c_binding
#include "finclude/petscdef.h"
implicit none

   interface  

     subroutine PetscViewerInitASCII_XML(viewer, ierr) bind(c,name='PetscViewerInitASCII_XMLRoutine')
     use iso_c_binding
     implicit none
     PetscViewer     :: viewer
     integer(c_int)  :: ierr
     end subroutine PetscViewerInitASCII_XML
 
   end interface

   contains

   subroutine PetscViewerXMLStartSection(viewer, elemname, desc, ierr)
     use iso_c_binding
     implicit none
     PetscViewer    :: viewer
     character(*)   :: elemname
     character(*)   :: desc
     integer(c_int) :: ierr
     character(len=len_trim(elemname)+1,kind=c_char) :: elemname_c
     character(len=len_trim(desc)+1,kind=c_char) :: desc_c
     interface
       subroutine PetscViewerXMLStartSectionRoutine(viewer, elemname, desc, ierr) bind(c,name='PetscViewerXMLStartSectionRoutine')
       use iso_c_binding
       implicit none
       PetscViewer     :: viewer
       character(kind=c_char) :: elemname(*)
       character(kind=c_char) :: desc(*)
       integer(c_int)  :: ierr
       end subroutine
     end interface

     elemname_c = trim(elemname)//c_null_char
     desc_c = trim(desc)//c_null_char

     call PetscViewerXMLStartSectionRoutine(viewer, elemname_c, desc_c, ierr)

   end subroutine

   subroutine PetscViewerXMLEndSection(viewer, elemname, ierr)
     use iso_c_binding
     implicit none
     PetscViewer    :: viewer
     character(*)   :: elemname
     integer(c_int) :: ierr
     character(len=len_trim(elemname)+1,kind=c_char) :: elemname_c
     interface
       subroutine PetscViewerXMLEndSectionRoutine(viewer, elemname, ierr) bind(c,name='PetscViewerXMLEndSectionRoutine')
       use iso_c_binding
       implicit none
       PetscViewer     :: viewer
       character(kind=c_char) :: elemname(*)
       integer(c_int)  :: ierr
       end subroutine
     end interface

     elemname_c = trim(elemname)//c_null_char

     call PetscViewerXMLEndSectionRoutine(viewer, elemname_c, ierr)

   end subroutine

   subroutine PetscViewerXMLPutString(viewer, elemname, desc, value, ierr)
     use iso_c_binding
     implicit none
     PetscViewer    :: viewer
     character(*)   :: elemname
     character(*)   :: desc
     character(*)   :: value
     integer(c_int) :: ierr
     character(len=len_trim(elemname)+1,kind=c_char) :: elemname_c
     character(len=len_trim(desc)+1,kind=c_char) :: desc_c
     character(len=len_trim(value)+1,kind=c_char) :: value_c
     interface
       subroutine PetscViewerXMLPutStringRoutine(viewer, elemname, desc, value, ierr) bind(c,name='PetscViewerXMLPutStringRoutine')
       use iso_c_binding
       implicit none
       PetscViewer     :: viewer
       character(kind=c_char) :: elemname(*)
       character(kind=c_char) :: desc(*)
       character(kind=c_char) :: value(*)
       integer(c_int)  :: ierr
       end subroutine
     end interface

     elemname_c = trim(elemname)//c_null_char
     desc_c = trim(desc)//c_null_char
     value_c = trim(value)//c_null_char

     call PetscViewerXMLPutStringRoutine(viewer, elemname_c, desc_c, value_c, ierr)

   end subroutine

   subroutine PetscViewerXMLPutInt(viewer, elemname, desc, value, ierr)
     use iso_c_binding
     implicit none
     PetscViewer    :: viewer
     character(*)   :: elemname
     character(*)   :: desc
     integer(c_int) :: value
     integer(c_int) :: ierr
     character(len=len_trim(elemname)+1,kind=c_char) :: elemname_c
     character(len=len_trim(desc)+1,kind=c_char) :: desc_c
     interface
       subroutine PetscViewerXMLPutIntRoutine(viewer, elemname, desc, value, ierr) bind(c,name='PetscViewerXMLPutIntRoutine')
       use iso_c_binding
       implicit none
       PetscViewer     :: viewer
       character(kind=c_char) :: elemname(*)
       character(kind=c_char) :: desc(*)
       integer(c_int) :: value
       integer(c_int) :: ierr
       end subroutine
     end interface

     elemname_c = trim(elemname)//c_null_char
     desc_c = trim(desc)//c_null_char

     call PetscViewerXMLPutIntRoutine(viewer, elemname_c, desc_c, value, ierr)

   end subroutine

   subroutine PetscViewerXMLPutDouble(viewer, elemname, desc, value, valueformat, ierr)
     use iso_c_binding
     implicit none
     PetscViewer    :: viewer
     character(*)   :: elemname
     character(*)   :: desc
     PetscLogDouble :: value
     character(*)   :: valueformat
     integer(c_int) :: ierr
     character(len=len_trim(elemname)+1,kind=c_char) :: elemname_c
     character(len=len_trim(desc)+1,kind=c_char) :: desc_c
     character(len=len_trim(valueformat)+1,kind=c_char) :: valueformat_c
     interface
       subroutine PetscViewerXMLPutDoubleRoutine(viewer, elemname, desc, value, valueformat, ierr) &
       bind(c,name='PetscViewerXMLPutDoubleRoutine')
       use iso_c_binding
       implicit none
       PetscViewer     :: viewer
       character(kind=c_char) :: elemname(*)
       character(kind=c_char) :: desc(*)
       PetscLogDouble :: value
       character(kind=c_char) :: valueformat(*)
       integer(c_int) :: ierr
       end subroutine
     end interface

     elemname_c = trim(elemname)//c_null_char
     desc_c = trim(desc)//c_null_char
     valueformat_c = trim(valueformat)//c_null_char

     call PetscViewerXMLPutDoubleRoutine(viewer, elemname_c, desc_c, value, valueformat_c, ierr)

   end subroutine

end module XMLViewerF

